import 'dart:async';
import 'dart:io';

import 'package:plexus/plexus.dart';

//import 'package:io/ansi.dart' hide AnsiCode, AnsiCodeType, ansiOutputEnabled;

/// Whether formatted ANSI output is enabled for [wrapWith] and [AnsiCode.wrap].
///
/// By default, returns `true` if both `stdout.supportsAnsiEscapes` and
/// `stderr.supportsAnsiEscapes` from `dart:io` are `true`.
///
/// The default can be overridden by setting the [Zone] variable [AnsiCode] to
/// either `true` or `false`.
///
/// [overrideAnsiOutput] is provided to make this easy.
bool get ansiOutputEnabled => Zone.current['AnsiFanci'] ?? (stdout.supportsAnsiEscapes && stderr.supportsAnsiEscapes);

/// Allows overriding [ansiOutputEnabled] to [enableAnsiOutput] for the code run
/// within [body].
T overrideAnsiOutput<T>(bool enableAnsiOutput, T body()) => runZoned(body, zoneValues: {'AnsiFanci': enableAnsiOutput});

const ansiEscapeLiteral = '\x1B';
const ansiEscapeForScript = '\\033';

String actName(AnsiCodeType type) => type.toString().split('.')[1];

/// The type of code represented by [AnsiCode].
class AnsiCodeType {
  const AnsiCodeType._(this.name);

  /// A foreground color.
  static const AnsiCodeType foreground = AnsiCodeType._('foreground');

  /// A style.
  static const AnsiCodeType style = AnsiCodeType._('style');

  /// A background color.
  static const AnsiCodeType background = AnsiCodeType._('background');

  /// No formatting.
  static const AnsiCodeType none = AnsiCodeType._('none');

  final String name;

  @override
  String toString() => 'AnsiType.$name';
}

/// Standard ANSI escape code for customizing terminal text output.
///
/// [Source](https://en.wikipedia.org/wiki/ANSI_escape_code#Colors)
class AnsiCode {
  const AnsiCode(this.name, this.type, this.code, {this.resetCode = 39});

  /// A description of this code.
  final String name;

  /// The type of code that is represented.
  final AnsiCodeType type;

  /// The numeric value associated with this code.
  @nullable
  final String code;

  /// The [AnsiCode] that resets this value.
  ///
  /// Must be specified in constructor if type is [AnsiCodeType.style].
  final int resetCode;

  /// Represents the resetValue escaped for use in terminal output.
  String get reset => code == null ? '' : '$ansiEscapeLiteral[${resetCode}m';

  /// Represents the resetValue as an unescaped literal suitable for scripts.
  String get resetForScript => code == null ? '' :  '$ansiEscapeForScript[${resetCode}m';

  String resetValue({bool forScript = false}) => forScript ? resetForScript : reset;

  /// Represents the value escaped for use in terminal output.
  String get escape => code == null ? '' :  '$ansiEscapeLiteral[${code}m';

  /// Represents the value as an unescaped literal suitable for scripts.
  String get escapeForScript => code == null ? '' : '$ansiEscapeForScript[${code}m';

  /// Returns the escape value for a script if forScript is true, otherwise returns the escape value for terminal output.
  String escapeValue({bool forScript = false}) {
    return forScript ? escapeForScript : escape;
  }

  /// Wraps [value] with the [escape] value for this code, followed by
  /// [resetAll].
  ///
  /// If [forScript] is `true`, the return value is an unescaped literal. The
  /// value of [ansiOutputEnabled] is also ignored.
  ///
  /// Returns `value` unchanged if
  ///   * [value] is `null` or empty
  ///   * both [ansiOutputEnabled] and [forScript] are `false`.
  ///   * [type] is [AnsiCodeType.reset]
  String wrap(String input, {bool forScript = false}) => input.isNotEmpty && code != null ? '${escapeValue(forScript: forScript)}$input${resetValue(forScript: forScript)}' : input;

  @override
  String toString() => '$name ${actName(type)} ($code)';

  static const lightBlack = AnsiCode('light black', AnsiCodeType.foreground, '90');
  static const lightRed = AnsiCode('light red', AnsiCodeType.foreground, '91');
  static const lightGreen = AnsiCode('light green', AnsiCodeType.foreground, '92');
  static const lightYellow = AnsiCode('light yellow', AnsiCodeType.foreground, '93');
  static const lightBlue = AnsiCode('light blue', AnsiCodeType.foreground, '94');
  static const lightMagneta = AnsiCode('light magneta', AnsiCodeType.foreground, '95');
  static const lightCyan = AnsiCode('light cyan', AnsiCodeType.foreground, '96');
  static const lightWhite = AnsiCode('light white', AnsiCodeType.foreground, '97');

  static const none = AnsiCode('none', AnsiCodeType.none, null);

  static const orange = AnsiCode('light orange', AnsiCodeType.foreground, '38;2;191;131;97');
  static const diamond = AnsiCode('light diamond', AnsiCodeType.foreground, '38;2;179;227;226');
}
