import 'dart:async';

import 'package:plexus/plexus.dart';

abstract class IHasShortMessage {
  /// The short message for this event. Should be less than 40 characters.
  String get message;
}

abstract class IHasTimestamp {
  DateTime get timestamp;
}

abstract class IHasLongMessage {
  /// The long message for this event. Should contain the same information as [IHasMessage#message] and then expand on it.
  @nullable
  String get longMessage;
}

abstract class IHasSource {
  Module get source;
}

abstract class IHasBlob<B> {
  /// Non-string message passed to Logger.
  @nullable
  B get blob;
}

abstract class IHasLevel {
  Level get level;
}

abstract class IHasError {
  /// Associated error (if any) when recording errors messages.
  @nullable
  Object get error;

  /// Associated stackTrace (if any) when recording errors messages.
  @nullable
  StackTrace get stackTrace;
}

abstract class IHasZone {
  /// Zone of the calling code which resulted in this LogRecord.
  Zone get zone;
}

abstract class IHasIsInit {
  /// Does this event take place during initialization
  bool get init;
}

/// An event which will be logged by a [Logger]'s individual handlers.
///
/// [source] will usually be set by a Logger.
/// Manually setting [timestamp] is discouraged.
abstract class Event<B> implements IHasSource, IHasShortMessage, IHasLongMessage, IHasBlob<B>, IHasTimestamp, IHasLevel, IHasError, IHasZone, IHasIsInit {
  Event(@nullable DateTime timestamp): timestamp = timestamp ?? DateTime.now();

  @override
  final DateTime timestamp;
}

/// An event which will be logged by a [Logger]'s individual handlers.
///
/// [source] will usually be set by a Logger.
/// Manually setting [timestamp] is discouraged.
class StdEvent<B> extends Event<B> {
  StdEvent(this.source, this.level, this.message, {this.longMessage, this.error, this.stackTrace, this.blob, @nullable Zone zone, this.init = false, @nullable DateTime timestamp}) : zone = zone ?? Zone.current, super(timestamp);

  @override
  final Module source;

  @override
  final Level level;

  @override
  final String message;

  @override
  @nullable
  final String longMessage;

  /// Non-string message passed to Logger.
  @override
  @nullable
  final B blob;

  /// Associated error (if any) when recording errors messages.
  @override
  @nullable
  final Object error;

  /// Associated stackTrace (if any) when recording errors messages.
  @override
  @nullable
  final StackTrace stackTrace;

  /// Zone of the calling code which resulted in this LogRecord.
  @override
  final Zone zone;

  /// Does this event take place during initialization
  @override
  final bool init;
}

class KernelEvent<B> extends Event<B> {
  KernelEvent(this.source, this.level, this.signal, {this.longMessage, this.error, this.stackTrace, this.blob, @nullable Zone zone, this.init = false, @nullable DateTime timestamp}) : zone = zone ?? Zone.current, super(timestamp);

  @override
  final Module source;

  @override
  final Level level;

  @override
  String get message => signal.toString();

  @override
  @nullable
  final String longMessage;

  /// Non-string message passed to Logger.
  @override
  @nullable
  final B blob;

  /// Associated error (if any) when recording errors messages.
  @override
  @nullable
  final Object error;

  /// Associated stackTrace (if any) when recording errors messages.
  @override
  @nullable
  final StackTrace stackTrace;

  /// Zone of the calling code which resulted in this LogRecord.
  @override
  final Zone zone;

  /// Does this event take place during initialization
  @override
  final bool init;

  final KernelSignal signal;
}

class AuthEvent<B> extends Event<B> {
  AuthEvent(this.source, this.level, this.message, this.signal, {this.longMessage, this.error, this.stackTrace, this.blob, @nullable Zone zone, this.init = false, @nullable DateTime timestamp}) : zone = zone ?? Zone.current, super(timestamp);

  @override
  final Module source;

  final Level level;

  @override
  final String message;

  @override
  @nullable
  final String longMessage;

  /// Non-string message passed to Logger.
  @override
  @nullable
  final B blob;

  /// Associated error (if any) when recording errors messages.
  @override
  @nullable
  final Object error;

  /// Associated stackTrace (if any) when recording errors messages.
  @override
  @nullable
  final StackTrace stackTrace;

  /// Zone of the calling code which resulted in this LogRecord.
  @override
  final Zone zone;

  /// Does this event take place during initialization
  @override
  final bool init;

  final AuthSignal signal;
}

typedef bool InputTest(List<String> context, String current);

class RequestInputEvent<B> extends Event<B> {
  RequestInputEvent(this.source, this.level, this.context, this.message, this.test, {this.longMessage, this.error, this.stackTrace, this.blob, @nullable Zone zone, this.init = false, @nullable DateTime timestamp}) : zone = zone ?? Zone.current, super(timestamp);

  @override
  final Module source;

  @override
  final Level level;

  /// This will be used for the input prompt. It should usually contain the context.
  @override
  final String message;

  @override
  @nullable
  final String longMessage;

  /// Non-string message passed to Logger.
  @override
  @nullable
  final B blob;

  /// Associated error (if any) when recording errors messages.
  @override
  @nullable
  final Object error;

  /// Associated stackTrace (if any) when recording errors messages.
  @override
  @nullable
  final StackTrace stackTrace;

  /// Zone of the calling code which resulted in this LogRecord.
  @override
  final Zone zone;

  /// Does this event take place during initialization
  @override
  final bool init;

  /// Contains all commands/arguments parsed prior to this.
  final List<String> context;

  /// Validates [current] input with access to [context] previous (valid) inputs.
  final InputTest test;

  bool testInput(String current) => test(context, current);
}
