import 'dart:async';

import 'package:plexus/plexus.dart';
import 'package:rxdart/rxdart.dart' show Observable, Subject;

/// A special StreamController that captures the latest item that has been
/// added to the controller, and emits that as the first item to any new
/// listener.
///
/// This subject allows sending data, error and done events to the listener.
/// The latest item that has been added to the subject will be sent to any
/// new listeners of the subject. After that, any new events will be
/// appropriately sent to the listeners. It is possible to provide a seed value
/// that will be emitted if no items have been added to the subject.
///
/// BehaviorSubject is, by default, a broadcast (aka hot) controller, in order
/// to fulfill the Rx Subject contract. This means the Subject's `stream` can
/// be listened to multiple times.
///
/// ### Example
///
///     final subject = new BehaviorSubject<int>();
///
///     subject.add(1);
///     subject.add(2);
///     subject.add(3);
///
///     subject.stream.listen(print); // prints 3
///     subject.stream.listen(print); // prints 3
///     subject.stream.listen(print); // prints 3
///
/// ### Example with seed value
///
///     final subject = new BehaviorSubject<int>.seeded(1);
///
///     subject.stream.listen(print); // prints 1
///     subject.stream.listen(print); // prints 1
///     subject.stream.listen(print); // prints 1
class StreamQueueObservable<T> extends Subject<T> implements Observable<T> {
  /// If passed, [seedValue] must be either a [T] or a [List<T>] and ]seedError] must be either a [ErrorBundle] or a [List<StackedError>]. If one is a List, the other must also be. If one is null (assuming they are [T] and [ErrorBundle]), the other must **not** be null. If they are Lists and one slot is null, the corresponding slot in the other List must **not** be null.
  factory StreamQueueObservable({
    @nullable
    VoidCallback onListen,
    @nullable
    VoidCallback onCancel,
    // T? seedValue,
    // ErrorBundle? seedError,
    // int capacity = BufferedQueue.MINIMUM_CAPACITY,
    bool sync = false,
  }) {
    // assert(!(seedValue != null && seedError != null), 'seedValue and seedError cannot both be non-null.');
    // assert(capacity > 0, 'maxEntries must be greater than 0');

    // ignore: close_sinks
    final controller = StreamController<T>.broadcast(
      onListen: onListen,
      onCancel: onCancel,
      sync: sync,
    );

    // final valueList = List<T?>(capacity);
    // valueList[0] = seedValue;

    // final errorList = List<ErrorBundle?>(capacity);
    // errorList[0] = seedError;

    //ignore: argument_type_not_assignable
    // final queue = QueuedObservable<T>.fromList(valueList, errorList);
    final queue = PatientStreamQueue<T>();

    return StreamQueueObservable<T>.internal(
      controller,
      Observable<T>(controller.stream),
      // Observable<T>.defer(() {
      //   if (queue.isError) {
      //     //ignore: argument_type_not_assignable
      //     scheduleMicrotask(() => controller.addError(queue.error!.error, queue.error!.stackTrace));
      //   } else if (queue.value != null) {
      //     return Observable<T>(controller.stream).startWith(queue.value);
      //   }

      //   return controller.stream;
      // }, reusable: true),
      queue,
    );
  }

  StreamQueueObservable.internal(
    StreamController<T> controller,
    Observable<T> observable,
    this.queue,
  ) : super(controller, observable);

  @override
  void onAdd(T event) => queue.add(event);

  @override
  void onAddError(Object error, [@nullable StackTrace stackTrace]) => queue.addError(error, stackTrace);

  @override
  final PatientStreamQueue<T> queue;
}
