import 'dart:async';

import 'package:collection/collection.dart' show ListEquality;
import 'package:meta/meta.dart';
import 'package:plexus/expose.dart';
import 'package:plexus/plexus.dart';

typedef Module ModuleBuilder();
typedef void EventHandler(Event event);

abstract class ILog {
  /// Automatically record stack traces for any message of this level or above.
  ///
  /// Because this is expensive, this is off by default.
  Level get recordStackTraceAtLevel;

  /// Effective level considering the levels established in this logger's
  /// parents.
  Level get level;

  /// Override the level for this particular [Module] and its children.
  set level(Level value);

  /// Whether a message for [value]'s level is loggable in this logger.
  bool isLoggable(Level value);

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  void log(Level logLevel, dynamic message, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, @nullable dynamic blob});

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  void logAuth(Level logLevel, dynamic message, AuthSignal signal, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, @nullable dynamic blob});

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  void logKernel(KernelSignal signal, {@nullable String longMessage, Level logLevel = Level.FINEST, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, @nullable dynamic blob});

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  void logRequestInput(List<String> context, String prompt, InputTest test, {@nullable String longMessage, Level logLevel = Level.FINEST, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, @nullable dynamic blob});
}

abstract class Module implements ILog {
  static Module get root => modules[[]];

  static final Map<List<String>, Module> modules = {[]: DefModule._([], Level.ALL, RootModule().add)};

  static M registerModule<M extends Module>(List<String> package, ModuleBuilder builder) {
    modules.putIfAbsent(package, builder);
    final ret = modules[package];
    return ret as M;
  }

  factory Module(List<String> package, {Level level = Level.INFO, @nullable EventHandler eventHandler}) => DefModule(package, level: level, eventHandler: eventHandler);

  factory Module.fromString(String package, {Level level = Level.INFO, @nullable EventHandler eventHandler}) => Module(packageFromString(package), level: level, eventHandler: eventHandler);

  Level get _level;
  void set _level(Level newLevel);

  UnmodifiableListView<String> get package;

  @protected
  EventHandler get eventHandler;

  String get immediateModule;
  Module get parent;

  String get name;

  /// The full name of this logger, which includes the parent's full name.
  String get fullName;

  String get packageifiedPackage;
  String get pathifiedPackage;
  String get pathifiedPackageWithExtension;

  //String get fullyQualifiedPackage => package.isNotEmpty ? '$prefix:$pathifiedPackage' : '/';

  static List<String> packageFromString(String package) => package.split(RegExp(r'\.\/'));
}

mixin HasModule implements Module {
  @protected
  @override
  EventHandler get eventHandler => module.eventHandler;

  @protected
  Module get module => Module(package, level: level, eventHandler: RootModule().add);

  @override
  String get immediateModule => module.immediateModule;
  
  @override
  Module get parent => module.parent;

  @override
  String get name => module.name;

  /// The full name of this logger, which includes the parent's full name.
  @override
  String get fullName => module.fullName;

  @override
  String get packageifiedPackage => module.packageifiedPackage;
  
  @override
  String get pathifiedPackage => module.pathifiedPackage;

  @override
  String get pathifiedPackageWithExtension => module.pathifiedPackageWithExtension;


  @override
  Level get recordStackTraceAtLevel => module.recordStackTraceAtLevel;


  @override
  Level get _level => module._level;

  @override
  void set _level(Level newLevel) => module._level = newLevel;

  /// Effective level considering the levels established in this logger's
  /// parents.
  @override
  Level get level => module.level;

  /// Override the level for this particular [Module] and its children.
  @override
  set level(Level value) => module.level = value;

  /// Whether a message for [value]'s level is loggable in this logger.
  @override
  bool isLoggable(Level value) => module.isLoggable(value);

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  @override
  void log(Level logLevel, dynamic message, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, @nullable dynamic blob}) => module.log(logLevel, message, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, init: init, blob: blob);

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  @override
  void logAuth(Level logLevel, dynamic message, AuthSignal signal, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) => module.logAuth(logLevel, message, signal, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, init: init, blob: blob);

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  @override
  void logKernel(KernelSignal signal, {@nullable String longMessage, Level logLevel = Level.FINEST, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) => module.logKernel(signal, logLevel: level, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, init: init, blob: blob);

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  @override
  void logRequestInput(List<String> context, String prompt, InputTest test, {@nullable String longMessage, @nullable Level logLevel = Level.FINEST, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) => module.logRequestInput(context, prompt, test, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, init: init, blob: blob);

  bool operator ==(other) => other is Module && ListEquality<String>().equals(package, other.package);

  @override
  String toString() => pathifiedPackage;

  @override
  int get hashCode => package.hashCode;
}

mixin ModuleMiddleware implements Module {
  /// Automatically record stack traces for any message of this level or above.
  ///
  /// Because this is expensive, this is off by default.
  @override
  final Level recordStackTraceAtLevel = Level.ERROR;

  @override
  String get immediateModule => package.isNotEmpty ? package.last : '/';
  
  @override
  Module get parent => Module(package.sublist(0, package.length));

  @override
  String get name => immediateModule;

  /// The full name of this logger, which includes the parent's full name.
  @override
  String get fullName => packageifiedPackage;

  @override
  String get packageifiedPackage => '${package.isNotEmpty ? package.join('.') : 'ROOT'}';
  
  @override
  String get pathifiedPackage => '/${package.isNotEmpty ? package.join('/') : ''}';

  @override
  String get pathifiedPackageWithExtension => '/${package.isNotEmpty ? '${package.join('/')}.dart' : ''}';

  /// Effective level considering the levels established in this logger's
  /// parents.
  @override
  Level get level => _level ?? parent.level ?? (_level = RootModule.defaultLevel);

  /// Override the level for this particular [Module] and its children.
  @override
  set level(Level value) => (value != null || parent != null) ? _level = value : _level;

  /// Whether a message for [value]'s level is loggable in this logger.
  @override
  bool isLoggable(Level value) => value >= level;

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  @override
  void log(Level logLevel, dynamic message, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, dynamic blob}) {
    if (message is Function) {
      message = message();
    }

    if (message is! String) {
      blob ??= message;
      message = message.toString();
    }

    if (stackTrace == null && logLevel >= recordStackTraceAtLevel) {
      stackTrace = StackTrace.current;
      error ??= 'autogenerated stack trace for $logLevel $message';
    }
    
    zone ??= Zone.current;

    var event = StdEvent(this, logLevel, message, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, blob: blob, init: init);

    eventHandler(event);
  }

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  @override
  void logAuth(Level logLevel, dynamic message, AuthSignal signal, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) {
    if (message is Function) {
      message = message();
    }

    if (!message is String) {
      blob ??= message;
      message = message.toString();
    }

    if (stackTrace == null && logLevel >= recordStackTraceAtLevel) {
      stackTrace = StackTrace.current;
      error ??= 'autogenerated stack trace for $logLevel $message';
    }
    
    zone ??= Zone.current;

    var event = AuthEvent(this, logLevel, message, signal, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, blob: blob, init: init);

    eventHandler(event);
  }

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  @override
  void logKernel(KernelSignal signal, {@nullable String longMessage, Level logLevel = Level.FINEST, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) {
    if (stackTrace == null && logLevel >= recordStackTraceAtLevel) {
      stackTrace = StackTrace.current;
      error ??= 'autogenerated stack trace for $logLevel $signal';
    }

    zone ??= Zone.current;

    var event = KernelEvent(this, logLevel, signal, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, blob: blob, init: init);

    eventHandler(event);
  }

  /// Adds a log event for a [message] at a particular [logLevel] if
  /// `isLoggable(logLevel)` is true.
  ///
  /// Use this method to create log entries for user-defined levels. To record a
  /// message at a predefined level (e.g. [Level.INFO], [Level.WARNING], etc)
  /// you can use their specialized methods instead (e.g. [info], [warning],
  /// etc).
  ///
  /// If [message] is a [Function], it will be lazy evaluated. Additionally, if
  /// [message] or its evaluated value is not a [String], then 'toString()' will
  /// be called on the object and the result will be logged. The log event will
  /// contain a field holding the original object.
  ///
  /// The log event will also contain a field for the zone in which this call
  /// was made. This can be advantageous if a log listener wants to handler
  /// records of different zones differently (e.g. group log records by HTTP
  /// request if each HTTP request handler runs in it's own zone).
  @override
  void logRequestInput(List<String> context, String prompt, InputTest test, {@nullable String longMessage, Level logLevel = Level.FINEST, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) {
    if (stackTrace == null && logLevel >= recordStackTraceAtLevel) {
      stackTrace = StackTrace.current;
      error ??= 'autogenerated stack trace for $logLevel $context "$prompt"';
    }
    zone ??= Zone.current;

    var event = RequestInputEvent(this, logLevel, context, prompt, test, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, blob: blob, init: init);

    eventHandler(event);
  }

  bool operator ==(other) => other is Module && packageifiedPackage == other.packageifiedPackage;

  @override
  String toString() => pathifiedPackage;

  @override
  int get hashCode => package.hashCode;
}

class DefModule with ModuleMiddleware {
  factory DefModule(List<String> package, {Level level = Level.INFO, @nullable EventHandler eventHandler}) => Module.registerModule<DefModule>(package, () => DefModule._(package, level, eventHandler));

  DefModule._(List<String> package, this._level, @nullable EventHandler eventHandler)
      : assert(package.every((module) => !module.contains(RegExp(r'[^a-z0-9_]'))), 'Modules can only consist of lowercase alphanumeric characters and underscores.'),
        assert(package.every((module) => module.startsWith(RegExp(r'[a-z]'))), 'Modules must start with a lowercase letter.'),
        package = UnmodifiableListView(package),
        eventHandler = eventHandler ?? voidEventHandler;

  factory DefModule.fromString(String package, {Level level = Level.INFO, @nullable EventHandler eventHandler}) => DefModule(Module.packageFromString(package), level: level, eventHandler: eventHandler);

  @override
  final UnmodifiableListView<String> package;

  @override
  final EventHandler eventHandler;

  @override
  Level _level;
}
