import 'dart:async';

import 'package:async/async.dart';
import 'package:intl/intl.dart';
import 'package:logging/logging.dart' as dlog;
import 'package:meta/meta.dart';
import 'package:plexus/expose.dart';
import 'package:plexus/plexus.dart';
import 'package:rxdart/rxdart.dart' show Observable;

// typedef Future<void> AsyncCallback();

final dtfmt = DateFormat('yyyy-MM-ddTHH-mm-ss.SSS');

/// There is a single [RootModule] instance, which should **only** be managed (setting level, etc.) by the final product (i.e. Flutter app (not library) or CLI (seperated from api)) and should be accessed statically outside this library.
class RootModule extends StreamQueueObservable<Event> with ModuleMiddleware {
  @nullable
  static RootModule _instance;

  static const Level defaultLevel = Level.ALL;

  factory RootModule({bool authEnabled = false, AsyncVoidCallback initHandler = asyncVoidHandler}) {
    final controller = StreamController<Event>.broadcast(onListen: () {}, onCancel: () {}, sync: false);
    return _instance ??= RootModule._(controller, Observable.defer(() => controller.stream), PatientStreamQueue<Event>(), authEnabled, initHandler);
  }

  RootModule._(StreamController<Event> controller, Observable<Event> observable, PatientStreamQueue<Event> queue, this.authEnabled, this.initHandler) : super.internal(controller, observable, queue);

  final bool authEnabled;
  final AsyncVoidCallback initHandler;

  @override
  Level _level = defaultLevel;

  @override
  set level(Level newLevel) {
    super.level = newLevel;
    dlog.Logger.root.level = level.dlogLevel;
  }

  @override
  final Level recordStackTraceAtLevel = Level.ERROR;

  @override
  final UnmodifiableListView<String> package = UnmodifiableListView([]);

  // void alert(dynamic error, [StackTrace stackTrace]) => tap.addError(error, stackTrace);

  final Map<String, StreamSubscription<Event>> subscriptions = {};

  /// Subscribes to events. You must unsubscribe using the [unsubscribe] method.
  StreamSubscription<Event> subscribe(String id, ValueCallback<Event> callback) {
    if (subscriptions.containsKey(id)) {
      log(Level.WARNING, 'Subscription with id `$id` already exists. Not added.');
      return subscriptions[id];
    } 
    
    final sub = stream.listen(callback);
    subscriptions[id] = sub;
    return sub;
  }

  Future<void> unsubscribe(String id) async {
    if (subscriptions.containsKey(id)) {
      await subscriptions[id].cancel();
      subscriptions.remove(id);
    }
  }

  Future<void> unsubscribeAll() async {
    final _stream = StreamController<String>();
    final _queue = StreamQueue<String>(_stream.stream);
    
    subscriptions.forEach((id, sub) {
      _stream.add(id);
    });

    while (await _queue.hasNext) {
      final _this = await _queue.next;
      await subscriptions[_this].cancel();
      subscriptions.remove(_this);
    } 
  }

  /// Returns a [StreamSubscription] of messages added to this [Logger] that pass [test].
  ///
  /// You can listen for messages using the standard stream APIs, for instance:
  ///
  /// ```dart
  /// logger.registerHandler((event) { ... });
  /// ```
  @optionalTypeArgs
  StreamSubscription<Event> watch(String id, ValueCallback<Event> callback, ValueTest<Event> test) {
    final _stream = stream.where(test);
    final sub = _stream.listen((e) => callback(e));
    subscriptions[id] = sub;
    return sub;
  }

  // @override
  StreamSubscription<E> watchType<E extends Event>(String id, ValueCallback<E> callback) {
    final _stream = stream.where((e) => e is E).cast<E>();
    final sub = _stream.listen((e) => callback(e));
    subscriptions[id] = sub;
    return sub;
  }

  Future<void> preInit() async {
    log(Level.INFO, 'Initializing Logger...', init: true);
    level = defaultLevel;
    dlog.Logger.root.onRecord.listen((record) {
      final shortMessage = record.message.length > 80 ? '${record.message.substring(0, 80 - 3)}...' : record.message;
      final longMessage = record.message.length > 80 ? record.message : null;
      add(StdEvent(
        Module(record.loggerName.toLowerCase().replaceAll(r'[^a-z0-9_\.]', '_').split('.')),
        Level.dlogToOcean[record.level],
        shortMessage,
        longMessage: longMessage,
        error: record.error,
        stackTrace: record.stackTrace,
        blob: record.object,
        timestamp: record.time,
        zone: record.zone,
      ));
    });
    subscribe('__console__', (record) {
      if (record.level >= record.source.level) {
        final timeFmt = AnsiCode.diamond.wrap(dtfmt.format(record.timestamp));
        final tag = record.level.wrappedName;
        final loggerName = AnsiCode.lightMagneta.wrap(record.source.fullName);
        var msg;
        if (record.level == Level.PRINT) {
          msg = record.message;

          // TODO(mpfaff): add `!` non-null assertion to [longMessage]
          if (record.longMessage != null && record.longMessage.isNotEmpty) {
            msg += '\n | ';
            // TODO(mpfaff): add `!` non-null assertion to [longMessage]
            final lines = record.longMessage.split('\n');
            while (lines.last.isEmpty) {
              lines.removeLast();
            }
            lines.forEach((l) => msg += '\n | $l');
            msg += '\n | ';
          }

          msg = record.level.ansiCode.wrap(msg);
        } else if (record.level == Level.ERROR || record.level == Level.FATAL) {
          final errorMsg = ' ${record.error ?? 'Auto-generated stacktrace'} |';
          msg = '${record.message}\n |$errorMsg';
          final stackList = (record.stackTrace ?? StackTrace.current).toString().split('\n');

          while (stackList.last.isEmpty) {
            stackList.removeLast();
          }

          stackList.forEach((l) => msg += '\n | $l');

          msg = '$msg\n |${'' * (errorMsg.length - 1)}|';

          msg = record.level.ansiCode.wrap(msg);

          msg = '[$timeFmt]: [$tag]: [$loggerName]: ' + msg;
        } else {
          msg = record.message;

          // TODO(mpfaff): add `!` non-null assertion to [longMessage]
          if (record.longMessage != null && record.longMessage.isNotEmpty) {
            msg += '\n | ';
            // TODO(mpfaff): add `!` non-null assertion to [longMessage]
            final lines = record.longMessage.split('\n');
            while (lines.last.isEmpty) {
              lines.removeLast();
            }
            lines.forEach((l) => msg += '\n | $l');
            msg += '\n | ';
          }

          msg = record.level.ansiCode.wrap(msg);

          msg = '[$timeFmt]: [$tag]: [$loggerName]: ' + msg;
        }

        print(msg);
      }
    });
    log(Level.INFO, 'Initialized logger', init: true);
  }

  Future<void> init() async {
    log(Level.INFO, 'Initializing Ocean...', init: true);

    await initHandler();

    log(Level.INFO, 'Initialized Ocean...', init: true);
  }

  Future<void> registerInternalHandlers() async {
    watchType<KernelEvent>('__kernel__', kernelEventHandler);

    if (authEnabled) watchType<AuthEvent>('__auth__', authEventHandler);

    logAuth(Level.FINEST, 'Initial State', AuthSignal.SignedOut);
  }

  void kernelEventHandler(KernelEvent event) {
    switch (event.signal) {
      case KernelSignal.Terminate:
        quit();
        break;
    }
  }

  void requestInputHandler(RequestInputEvent event) {
    final command = echoPrompt(prompt: '');
    if (command.contains(r'[^a-z0-9-]')) {
      log(Level.ERROR, 'Command must be lowercase alphanumeric with dashes.');
      logRequestInput([], '', (_, __) => true);
    } else {
      // TODO(PfaffDev): Change RequestInput to its own event and add a context field of type [List<String>] containing the previous commands/arguments
    }
  }

  void authEventHandler(AuthEvent event) async {
    @nullable Object _error;
    @nullable StackTrace _stackTrace;

    if (event.signal == AuthSignal.SignedOut || event.signal == AuthSignal.SignInFailed) {
      _stackTrace = event.stackTrace ?? StackTrace.current;
      _error = event.error; //?? 'Sorry, no error provided.';
    }

    switch (event.signal) {
      case AuthSignal.SignedOut:
        log(Level.ERROR, 'Sign In Failed.', longMessage: 'Sign in failed for some reason. This is the default SignInFailed handler. Maybe? try again.', error: _error, stackTrace: _stackTrace);
        break;
      case AuthSignal.SignInFailed:
        log(Level.ERROR, 'Sign In Failed.', longMessage: 'Sign in failed for some reason. This is the default SignInFailed handler. Maybe? try again.', error: _error, stackTrace: _stackTrace);
        break;
      case AuthSignal.SignedIn:
        log(Level.SUCCESS, 'We\'re Signed In!');
        break;
      case AuthSignal.SignOutFailed:
        break;
    }
  }

  Future<void> start() async {
    await preInit();
    await runZoned<Future<void>>(
      () async {
        await init();

        print('Welcome to Deep Storage!');

        await registerInternalHandlers();
      },
      zoneSpecification: ZoneSpecification(
        print: (_, __, zone, message) => log(Level.PRINT, message, zone: zone),
        handleUncaughtError: (_, __, zone, error, stackTrace) => log(Level.ERROR, 'Error in Ocean Zone.', error: error, stackTrace: stackTrace, zone: zone),
      ),
    );
  }

  Future<void> quit() async {
    print('Termination in progress...');

    //TODO(mpfaff): Run quit handler

    print('Goodbye.');
  }

  @override
  EventHandler get eventHandler => this.add;
}
